import 'dart:convert';
import 'dart:io';

import 'package:http/http.dart' as http;
import 'package:mason_logger/mason_logger.dart';
import 'package:path/path.dart';

import '../exit.dart';
import 'auth.dart';
import 'metadata.dart';

class ApiException implements Exception {
  ApiException._(this.statusCode, this.message);

  final int statusCode;
  final String message;
}

class GlobeApi {
  GlobeApi({
    required this.metadata,
    required this.auth,
    required this.logger,
  });

  Map<String, String> get _headers {
    final currentSession = auth.currentSession;

    return {
      'X-Globe-Platform': 'globe_cli',
      if (currentSession != null)
        'Authorization': 'Bearer ${currentSession.jwt}',
    };
  }

  final GlobeMetadata metadata;
  final GlobeAuth auth;
  final Logger logger;

  /// Creates a [Uri] from the given [path], using the [GlobeMetadata.endpoint] as the base.
  Uri _buildUri(String path) {
    assert(path.startsWith('/'), 'path must start with a /.');
    return Uri.parse('${metadata.endpoint}/api$path');
  }

  /// Handles an [http.Response], throwing an [ApiException] if the response
  /// is not successful.
  ///
  /// If the response returns invalid JSON, a [FormatException] will be thrown.
  Object? _handleResponse(http.Response response) {
    // Let the decode throw on invalid json.
    final json = jsonDecode(response.body) as Map<String, Object?>;

    if (response.statusCode == 200) return json['data'];

    throw ApiException._(response.statusCode, json['message']! as String);
  }

  /// Throws an exception if the user is not authenticated.
  void requireAuth() {
    if (auth.currentSession == null) {
      throw Exception('The user is not authenticated.');
    }
  }

  Future<Settings> getSettings({
    required String orgId,
    required String projectId,
  }) async {
    requireAuth();

    final uri = _buildUri('/orgs/$orgId/projects/$projectId/settings');
    logger.detail('API Request: GET $uri');
    final responseJson = _handleResponse(
      await http.get(uri, headers: _headers),
    )! as Map<String, Object?>;

    return Settings.fromJson(responseJson);
  }

  /// Gets all of the organizations that the current user is a member of.
  Future<List<Organization>> getOrganizations() async {
    requireAuth();
    logger.detail('API Request: GET /user/orgs');
    final response = _handleResponse(
      await http.get(_buildUri('/user/orgs'), headers: _headers),
    )! as List<Object?>;

    return response
        .cast<Map<String, Object?>>()
        .map(Organization.fromJson)
        .toList();
  }

  /// Gets all of the projects that the current user is a member of.
  Future<List<Project>> getProjects({
    required String org,
  }) async {
    requireAuth();
    logger.detail('API Request: GET /orgs/$org/projects');
    final response = _handleResponse(
      await http.get(_buildUri('/orgs/$org/projects'), headers: _headers),
    )! as List<Object?>;
    return response.cast<Map<String, Object?>>().map(Project.fromJson).toList();
  }

  /// Creates a new project and returns it.
  Future<Project> createProject({
    required String orgId,
    required String name,
  }) async {
    logger.detail('API Request: POST /orgs/$orgId/projects');
    final request = http.Request('POST', _buildUri('/orgs/$orgId/projects'));
    request.headers.addAll(_headers);

    request.body = jsonEncode({'name': name});

    final response = _handleResponse(
      await request.send().then(http.Response.fromStream),
    )! as Map<Object?, Object?>;

    return Project.fromJson(response);
  }

  /// Deploy the current project to the given project.
  Future<String> deploy({
    required String orgId,
    required String projectId,
    required DeploymentEnvironment environment,
    required List<int> archive,
  }) async {
    requireAuth();
    logger.detail('API Request: POST /orgs/$orgId/projects/$projectId/deploy');
    final request = http.MultipartRequest(
      'POST',
      _buildUri('/orgs/$orgId/projects/$projectId/deploy'),
    );

    request.headers.addAll(_headers);
    request.fields['environment'] = environment.name;

    // Check that the archive is not too large.
    // We devide by 8 because the archive is a list of bytes not bits.
    if (archive.length > 40 * 1000 * 1000 / 8) {
      throw Exception('The project must be at most 40mb.');
    }

    request.files.add(
      http.MultipartFile.fromBytes(
        'projectArchive',
        archive,
        filename: 'archive.zip',
      ),
    );

    final response = _handleResponse(
      await request.send().then(http.Response.fromStream),
    )! as Map<Object?, Object?>;

    return response['id']! as String;
  }
}

class Settings {
  Settings._({
    required this.entrypoint,
    required this.buildCommand,
    required this.rootDirectory,
    required this.connection,
  });

  factory Settings.fromJson(Map<String, dynamic> json) {
    switch (json) {
      case {
          'entrypoint': final String entrypoint,
          'buildCommand': final String buildCommand,
          'rootDirectory': final String rootDirectory,
          'connection': final Map<String, Object?>? rawConnection,
        }:
        final connection = rawConnection == null
            ? null
            : SettingsConnection.fromJson(rawConnection);

        /// If the project is linked with a git repository, "rootDirectory"
        /// is relative to the root of the git repository.
        /// Otherwise it is relative to the current directory.
        final projectRootDir = connection != null
            ? _findAnclosingGitDirectory(Directory.current)
            : Directory.current;

        final rootDir = Directory(join(projectRootDir.path, rootDirectory));

        return Settings._(
          entrypoint: File(join(rootDir.path, entrypoint)),
          buildCommand: buildCommand,
          rootDirectory: rootDir,
          connection: connection,
        );
      case _:
        throw const FormatException();
    }
  }

  static Directory _findAnclosingGitDirectory(Directory directory) {
    if (Directory(join(directory.path, '.git')).existsSync()) return directory;

    if (directory.path == directory.parent.path) {
      throw Exception('Could not find a git directory.');
    }

    return _findAnclosingGitDirectory(directory.parent);
  }

  final File entrypoint;
  final String buildCommand;
  final Directory rootDirectory;
  final SettingsConnection? connection;

  void validate({required Logger logger}) {
    if (!rootDirectory.existsSync()) {
      logger.err(
        'The root directory for this project does not exist: ${rootDirectory.path}',
      );
      exitOverride(ExitCode.software.code);
    }
    if (!File(join(rootDirectory.path, 'pubspec.yaml')).existsSync()) {
      logger.err('No pubspec.yaml found at: ${rootDirectory.path}');
      exitOverride(ExitCode.software.code);
    }
    if (!entrypoint.existsSync()) {
      logger.err(
        "The project's entrypoint does not exist: ${entrypoint.path}",
      );
      exitOverride(ExitCode.software.code);
    }
  }
}

class SettingsConnection {
  SettingsConnection._({
    required this.owner,
    required this.repository,
    required this.branch,
  });

  factory SettingsConnection.fromJson(Map<String, dynamic> json) {
    switch (json) {
      case {
          'owner': final String owner,
          'repository': final String repository,
          'branch': final String branch,
        }:
        return SettingsConnection._(
          owner: owner,
          repository: repository,
          branch: branch,
        );
      case _:
        throw const FormatException();
    }
  }

  final String owner;
  final String repository;
  final String branch;
}

class Organization {
  Organization._({
    required this.id,
    required this.name,
    required this.slug,
    required this.type,
    // required this.role,
    required this.createdAt,
    required this.updatedAt,
  });

  factory Organization.fromJson(Map<dynamic, dynamic> json) {
    return switch (json) {
      {
        'id': final String id,
        'name': final String name,
        'type': final String type,
        // 'role': final String role,
        'slug': final String slug,
        'createdAt': final String createdAt,
        'updatedAt': final String updatedAt,
      } =>
        Organization._(
          id: id,
          name: name,
          slug: slug,
          type: OrganizationType.fromString(type),
          // role: Role.fromString(role),
          createdAt: DateTime.parse(createdAt),
          updatedAt: DateTime.parse(updatedAt),
        ),
      _ => throw const FormatException(),
    };
  }

  final String id;
  final String name;
  final String slug;
  final OrganizationType type;
  // final Role role;
  final DateTime createdAt;
  final DateTime updatedAt;
}

class Project {
  Project._({
    required this.id,
    required this.orgId,
    required this.slug,
    required this.createdAt,
    required this.updatedAt,
  });

  factory Project.fromJson(Map<dynamic, dynamic> json) {
    return switch (json) {
      {
        'id': final String id,
        'organizationId': final String organizationId,
        'slug': final String slug,
        'createdAt': final String createdAt,
        'updatedAt': final String updatedAt,
      } =>
        Project._(
          id: id,
          orgId: organizationId,
          slug: slug,
          createdAt: DateTime.parse(createdAt),
          updatedAt: DateTime.parse(updatedAt),
        ),
      _ => throw const FormatException(),
    };
  }

  final String id;
  final String orgId;
  final String slug;
  final DateTime createdAt;
  final DateTime updatedAt;
}

enum DeploymentEnvironment {
  preview,
  production;
}

enum Role {
  admin,
  basicMember;

  static Role fromString(String role) {
    switch (role) {
      case 'admin':
        return Role.admin;
      case 'basic_member':
        return Role.basicMember;
      default:
        throw ArgumentError.value(role, 'role');
    }
  }
}

enum OrganizationType {
  personal,
  organization;

  static OrganizationType fromString(String type) {
    switch (type) {
      case 'personal':
        return OrganizationType.personal;
      case 'organization':
        return OrganizationType.organization;
      default:
        throw ArgumentError.value(type, 'type');
    }
  }
}
